#include "VariableArray.h"
#include <iostream>
#include <vector>
#include <map>
#include <cstdlib>
using namespace std;

typedef unsigned int uint;

class Coord
{
	public:
		uint x; uint y;

		Coord() : x(0), y(0) {}

		Coord(uint coordX, uint coordY) : x(coordX), y(coordY) {} 

		bool operator == (Coord d) { return ((x == d.x) && (y == d.y)); }	
};

template< class Heuristika > class Node : public Heuristika
{
	private:
		void Swap(Coord c1, Coord c2);
		Coord& FindEmpty();
	public:
		Node();
		Node(Node<Heuristika>& n); 
		Node(Node<Heuristika>* n);

		VariableArray<Node<Heuristika>*> children;
		void GenChildren();


		friend ostream& operator << (ostream& s, Node<Heuristika>& n)
		{
			for (uint i = 0; i < 3; i++)
			{
				for (uint j = 0; j < 3; j++)
				{ cout << "[" << n._pole[j][i] << "]"; }
				cout << endl;
			}

			return s;
		} 
};

/** \brief Defaultni konstruktor

	Random vyplni pole hodnot.

**/
template<class Heuristika> Node<Heuristika>::Node()
{
	bool assigned[9];
	for (uint i = 0; i < 9; i++) 
	{ assigned[i] = false; }

	for (uint i = 0; i < 3; i++)
		for (uint j = 0; j < 3; j++)
		{
			uint value = rand()%9;
			while (assigned[value]) 
			{ value = rand()%9; }

			assigned[value] = true;
			this->_pole[j][i] = value;
		}	
}

/// \brief Kopirovaci konstruktor
template<class Heuristika> Node<Heuristika>::Node(Node<Heuristika>& n)
{
	for (uint i = 0; i < 3; i++)
		for (uint j = 0; j < 3; j++)
		{ this->_pole[j][i] = n._pole[j][i]; }
}

/// \brief Kopirovaci konstruktor
template<class Heuristika> Node<Heuristika>::Node(Node<Heuristika>* n)
{
	for (uint i = 0; i < 3; i++)
		for (uint j = 0; j < 3; j++)
		{ this->_pole[j][i] = n->_pole[j][i]; }
}

template<class Heuristika> void Node<Heuristika>::Swap(Coord c1, Coord c2)
{
	uint tmp = this->_pole[c1.x][c1.y];
	this->_pole[c1.x][c1.y] = this->_pole[c2.x][c2.y];
	this->_pole[c2.x][c2.y] = tmp;
}

template<class Heuristika> Coord& Node<Heuristika>::FindEmpty()
{
	Coord result;
	
	for (uint i = 0; i < 3; i++)
		for (uint j = 0; j < 3; j++)
		{
			if (this->_pole[j][i] == 0)
			{ result.x = j; result.y = i; return result; }
		}

	return result;
}

template<class Heuristika> void Node<Heuristika>::GenChildren()
{
	uint countChildren = 4;
	Coord empty = FindEmpty();

	if ((empty.x == 0) || (empty.x == 2)) { countChildren--; }
	if ((empty.y == 0) || (empty.y == 2)) { countChildren--; }

	children.Init(countChildren);
	uint iterator = 0;

	if (empty.x != 0) 
	{ 
		children[iterator] = new Node(this);
		children[iterator]->Swap(empty,Coord(empty.x-1,empty.y));	

		iterator++;
	} // and move left

	if (empty.x != 2) 
	{ 
		children[iterator] = new Node(this);
		children[iterator]->Swap(empty,Coord(empty.x+1,empty.y));

		iterator++;
	} // and move right

	if (empty.y != 0) 
	{ 
		children[iterator] = new Node(this);
		children[iterator]->Swap(empty,Coord(empty.x,empty.y-1));	
		iterator++; 
	} // and move top

	if (empty.y != 2) 
	{ 
		children[iterator] = new Node(this); 
		children[iterator]->Swap(empty,Coord(empty.x,empty.y+1));
		iterator++; 
	} // and move bottom		
}


class Layout
{
	protected:
		map<uint,Coord> _pos;
	public:
		Layout()
		{
			_pos.insert(make_pair(1,Coord(0,0)));
			_pos.insert(make_pair(2,Coord(1,0)));
			_pos.insert(make_pair(3,Coord(2,0)));
			_pos.insert(make_pair(4,Coord(2,1)));
			_pos.insert(make_pair(5,Coord(2,2)));
			_pos.insert(make_pair(6,Coord(1,2)));
			_pos.insert(make_pair(7,Coord(0,2)));
			_pos.insert(make_pair(8,Coord(0,1)));
		}
};

class HamiltonDistance : public Layout
{
	protected:
		uint _pole[3][3];

	public:
		uint Value()
		{
			uint result = 0;

			for (uint i = 0; i < 3; i++)
			for (uint j = 0; j < 3; j++)
			{
				if (_pole[j][i] != 0)
				{
					result += abs((int)(_pos[_pole[j][i]].x - j)) + abs((int)(_pos[_pole[j][i]].y - i));
				}
			}
			return result;
		}
};

class FixedItems : public Layout
{
	protected:
		uint _pole[3][3];

	public:
		FixedItems() : Layout() {}


		uint Value()
		{
			uint result = 8;

			for (uint i = 0; i < 3; i++) for (uint j = 0; j < 3; j++)
			{
				Coord tmp; tmp.x = j; tmp.y = i;

/*		for (uint a = 0; a < 3; a++)
				{
					for (uint b = 0; b < 3; b++)
					{
						cout << "[" << _pole[b][a] << "]";
					}
					cout << endl;
				}

				cout << endl;

//		if (tmp == this->_pos[this->_pole[j][i]])
//		{
//			cout << "X: " << j << " Y: " << i << " " <<  this->_pole[j][i] << " is on correct place" << endl;
//			result--; 
//		}
*/}

			return result;
		}
};

#include <cstdlib>
#include <ctime>


#define TRYNUM 10
#define MAXCYCLE 30

int main()
{
	srand((unsigned)time(NULL));

	Node<FixedItems>* initialFixed = new Node<FixedItems>;
	Node<FixedItems>* currentFixed = initialFixed;

	
	cout << "Prezentace heuristik hamiltonovska vzdalenost a pocet spravne umistenych." << endl;

	for (uint i = 0; i < TRYNUM; i++)
	{
		Node<HamiltonDistance>* initialHamilton = new Node<HamiltonDistance>;
		Node<HamiltonDistance>* currentHamilton = initialHamilton;

		cout << "Hamiltonovska vzdalenost, pokus [" << i+1 << "/" << TRYNUM << "]" << endl;
		cout << "Pocatecni rozestaveni: " << endl << *currentHamilton << endl;

		uint counter = 0;

		while ((currentHamilton->Value() != 0) && (counter < MAXCYCLE))
		{
			currentHamilton->GenChildren();
			
			uint bestValue = currentHamilton->children[0]->Value();
			uint best = 0;

			//find the best child
			for (uint j = 1; j < currentHamilton->children.Length(); j++)
			{
				uint val = currentHamilton->children[j]->Value();
				if (val > bestValue) { bestValue = val; best = j; }
			}

			currentHamilton = currentHamilton->children[best];
		//	cout << *currentHamilton << endl;
			counter++;
		}

		if (counter < MAXCYCLE)
		{ cout << "Heuristika hamiltonovska vzdalenost uspesne nasla reseni po " << counter << " krocich." << endl; }
		else { cout << "Heuristika selhala. Po " << counter << " krocich nebyl nalezen cil." << endl;
			cout << "Aktualni rozestaveni: " << endl << *currentHamilton << endl; }
	}



	for (uint i = 0; i < TRYNUM; i++)
	{
		Node<FixedItems>* initialFixed = new Node<FixedItems>;
		Node<FixedItems>* currentFixed = initialFixed;

		cout << "Pocet spravne umistenych, pokus [" << i+1 << "/" << TRYNUM << "]" << endl;
		cout << "Pocatecni rozestaveni: " << endl <<*currentFixed << endl;

		uint counter = 0;

		while ((currentFixed->Value() != 0) && (counter < MAXCYCLE))
		{
			currentFixed->GenChildren();
			
			uint bestValue = currentFixed->children[0]->Value();
			uint best = 0;

			//find the best child
			for (uint j = 1; j < currentFixed->children.Length(); j++)
			{
				uint val = currentFixed->children[j]->Value();
				if (val > bestValue) { bestValue = val; best = j; }
			}

			currentFixed = currentFixed->children[best];
//			cout << *currentFixed << endl;
			counter++;
		}

		if (counter < MAXCYCLE)
		{ cout << "Heuristika pocet spravne umistenych uspesne nasla reseni po " << counter << " krocich." << endl; }
		else { cout << "Heuristika selhala. Po " << counter << " krocich nebyl nalezen cil." << endl;
			cout << "Aktualni rozestaveni: " << endl << *currentFixed << endl; }
	}	


}
